@@ -26,7 +26,8 @@ from mercurial.node import short as hgshort
 from mercurial.bookmarks import pushbookmark
 from mercurial.scmutil import revsingle
 
-from .util import log, die, output, git_to_hg_spaces, hgmode, branch_tip
+from .util import (die, output, git_to_hg_spaces, hgmode, branch_tip,
+    ref_to_name_reftype, BRANCH, BOOKMARK, TAG, user_config)
 
 
 class GitExporter(object):
@@ -47,6 +48,7 @@ class GitExporter(object):
         self.parser = parser
         self.processed_marks = set()
         self.processed_nodes = []
+        self.hgrc = user_config()
 
     def process(self):
         self.marks.store()  # checkpoint
@@ -59,24 +61,27 @@ class GitExporter(object):
                 die('unhandled command: %s' % line)
             getattr(self, 'do_%s' % command)()
 
+        updated_refs = {}
         for ref, node in self.parsed_refs.iteritems():
-            if ref.startswith('refs/heads/branches'):
-                branch = ref[len('refs/heads/branches'):]
-                if git_to_hg_spaces(branch) not in self.hgremote.branches:
-                    new_branch = True
-            elif ref.startswith('refs/heads/'):
-                bookmark = ref[len('refs/heads/'):]
-                old = self.hgremote.bookmarks.get(bookmark)
-                old = old.hex() if old else ''
-                if not pushbookmark(self.repo, bookmark, old, node):
-                    continue
-                push_bookmarks.append((bookmark, old, hghex(node)))
-            elif ref.startswith('refs/tags/'):
-                self.write_tag(ref)
-            else:
-                # transport-helper/fast-export bugs
-                log("Fast-export unexpected ref: %s" % ref, "WARNING")
+            if ref.startswith(self.hgremote.prefix):
+                # This seems to be a git fast-export bug
                 continue
+            name, reftype = ref_to_name_reftype(ref)
+            name = git_to_hg_spaces(name)
+            if reftype == BRANCH:
+                if name not in self.hgremote.branches:
+                    new_branch = True
+            elif reftype == BOOKMARK:
+                old = self.hgremote.bookmarks.get(name)
+                old = old.hex() if old else ''
+                if not pushbookmark(self.repo, name, old, node):
+                    continue
+                push_bookmarks.append((name, old, hghex(node)))
+            elif reftype == TAG:
+                self.write_tag(name, node)
+            else:
+                assert False, "unexpected reftype: %s" % reftype
+            updated_refs[ref] = node
 
         success = False
         try:
@@ -96,9 +101,20 @@ class GitExporter(object):
                 die("unknown hg exception: %s" % e)
         # TODO: handle network/other errors?
 
-        for ref in self.parsed_refs:
+        for ref, node in updated_refs.items():
             if success:
-                output("ok %s" % ref)
+                status = ""
+                name, reftype = ref_to_name_reftype(ref)
+                gitify_ref = self.hgremote.make_gitify_ref(name, reftype)
+                last_known_rev = self.marks.tips.get(gitify_ref)
+                new_rev = self.repo[node].rev()
+                if last_known_rev is not None and last_known_rev == new_rev:
+                    # up to date status tells git that nothing has changed
+                    # during the push for this ref, which prevents it from
+                    # printing pointless status info to the user such as:
+                    #  * [new branch]      master -> master
+                    status = " up to date"
+                output("ok %s%s" % (ref, status))
             else:
                 output("error %s non-fast forward" % ref)  # TODO: other errors as well
         output()
@@ -190,9 +206,9 @@ class GitExporter(object):
                         self.repo[parent_from].manifest():
                     files[file] = {'ctx': self.repo[parent_from][file]}
 
-        if ref.startswith('refs/heads/branches/'):
-            extra['branch'] = git_to_hg_spaces(
-                ref[len('refs/heads/branches/'):])
+        name, reftype = ref_to_name_reftype(ref)
+        if reftype == BRANCH:
+            extra['branch'] = git_to_hg_spaces(name)
 
         def get_filectx(repo, memctx, file):
             filespec = files[file]
@@ -230,9 +246,7 @@ class GitExporter(object):
     def do_feature(self):
         pass  # Ignore
 
-    def write_tag(self, ref):
-        node = self.parsed_refs[ref]
-        tag = git_to_hg_spaces(ref[len('refs/tags/'):])
+    def write_tag(self, name, node):
         branch = self.repo[node].branch()
         # Calling self.repo.tag() doesn't append the tag to the correct
         # commit. So I copied some of localrepo._tag into here.
@@ -248,7 +262,7 @@ class GitExporter(object):
         if old_tags and old_tags[-1] != '\n':
             newtags.append('\n')
 
-        encoded_tag = encoding.fromlocal(tag)
+        encoded_tag = encoding.fromlocal(name)
         tag_line = '%s %s' % (hghex(node), encoded_tag)
         if tag_line in old_tags:
             return  # Don't commit a tag that was previously committed
@@ -257,14 +271,14 @@ class GitExporter(object):
         def get_filectx(repo, memctx, file):
             return memfilectx(file, ''.join(newtags))
 
-        if tag in self.parsed_tags:
-            author, message = self.parsed_tags[tag]
+        if name in self.parsed_tags:
+            author, message = self.parsed_tags[name]
             user, date, tz = author
             date_tz = (date, tz)
         else:
-            message = "Added tag %s for changeset %s" % (tag, hgshort(node))
-            user = None
-            date_tz = None
+            message = "Added tag %s for changeset %s" % (name, hgshort(node))
+            user = self.hgrc.get("ui", "username", None)
+            date_tz = None  # XXX insert current date here
         ctx = memctx(self.repo,
             (branch_tip(self.repo, branch), self.NULL_PARENT), message,
             ['.hgtags'], get_filectx, user, date_tz, {'branch': branch})
